from typing import Any, Callable, Optional

import flask
from flask import Blueprint, abort, current_app, jsonify, make_response, request
from pydantic import ValidationError

from spectree._pydantic import (
    SerializedPydanticResponse,
    is_partial_base_model_instance,
    serialize_model_instance,
)
from spectree._types import ModelType
from spectree.plugins.base import Context, validate_response
from spectree.plugins.werkzeug_utils import WerkzeugPlugin, flask_response_unpack
from spectree.response import Response
from spectree.utils import cached_type_hints, get_multidict_items


class FlaskPlugin(WerkzeugPlugin):
    def get_current_app(self):
        return current_app

    def is_app_response(self, resp):
        return isinstance(resp, flask.Response)

    @staticmethod
    def is_blueprint(app: Any) -> bool:
        return isinstance(app, Blueprint)

    def request_validation(self, request, query, json, form, headers, cookies):
        """
        req_query: werkzeug.datastructures.ImmutableMultiDict
        req_json: dict
        req_headers: werkzeug.datastructures.EnvironHeaders
        req_cookies: werkzeug.datastructures.ImmutableMultiDict
        """
        req_query = get_multidict_items(request.args, query)
        req_headers = dict(iter(request.headers)) or {}
        req_cookies = get_multidict_items(request.cookies)
        has_data = request.method not in ("GET", "DELETE")
        # flask Request.mimetype is already normalized
        use_json = json and has_data and request.mimetype not in self.FORM_MIMETYPE
        use_form = form and has_data and request.mimetype in self.FORM_MIMETYPE

        request.context = Context(
            query.model_validate(req_query) if query else None,
            json.model_validate(request.get_json(silent=True) or {})
            if use_json
            else None,
            form.model_validate(self.fill_form(request)) if use_form else None,
            headers.model_validate(req_headers) if headers else None,
            cookies.model_validate(req_cookies) if cookies else None,
        )

    def validate_response(
        self,
        resp,
        resp_model: Optional[Response],
        skip_validation: bool,
        force_resp_serialize: bool,
    ):
        resp_validation_error = None
        payload, status, additional_headers = flask_response_unpack(resp)

        if self.is_app_response(payload):
            resp_status, resp_headers = payload.status_code, payload.headers
            payload = payload.get_data()
            # the inner flask.Response.status_code only takes effect when there is
            # no other status code
            if status == 200:
                status = resp_status
            # use the `Header` object to avoid deduplicated by `make_response`
            resp_headers.extend(additional_headers)
            additional_headers = resp_headers

        if not skip_validation and resp_model:
            try:
                response_validation_result = validate_response(
                    validation_model=resp_model.find_model(status),
                    response_payload=payload,
                    force_serialize=force_resp_serialize,
                )
            except ValidationError as err:
                errors = err.errors(include_context=False)
                response = make_response(errors, 500)
                resp_validation_error = err
            else:
                response = make_response(
                    self.get_current_app().response_class(
                        response_validation_result.payload.data,
                        mimetype="application/json",
                    )
                    if isinstance(
                        response_validation_result.payload,
                        SerializedPydanticResponse,
                    )
                    else response_validation_result.payload,
                    status,
                    additional_headers,
                )
        else:
            if is_partial_base_model_instance(payload):
                payload = self.get_current_app().response_class(
                    serialize_model_instance(payload).data,
                    mimetype="application/json",
                )
            response = make_response(payload, status, additional_headers)

        return response, resp_validation_error

    def validate(
        self,
        func: Callable,
        query: Optional[ModelType],
        json: Optional[ModelType],
        form: Optional[ModelType],
        headers: Optional[ModelType],
        cookies: Optional[ModelType],
        resp: Optional[Response],
        before: Callable,
        after: Callable,
        validation_error_status: int,
        skip_validation: bool,
        force_resp_serialize: bool,
        *args: Any,
        **kwargs: Any,
    ):
        response, req_validation_error = None, None
        if not skip_validation:
            try:
                self.request_validation(request, query, json, form, headers, cookies)
            except ValidationError as err:
                req_validation_error = err
                errors = err.errors(include_context=False)
                response = make_response(jsonify(errors), validation_error_status)

        before(request, response, req_validation_error, None)

        if req_validation_error is not None:
            assert response  # make mypy happy
            abort(response)

        if self.config.annotations:
            annotations = cached_type_hints(func)
            for name in ("query", "json", "form", "headers", "cookies"):
                if annotations.get(name):
                    kwargs[name] = getattr(
                        getattr(request, "context", None), name, None
                    )

        result = func(*args, **kwargs)

        response, resp_validation_error = self.validate_response(
            result,
            resp,
            skip_validation,
            force_resp_serialize,
        )
        after(request, response, resp_validation_error, None)

        return response



from typing import Dict

ONLINE_PAGE_TEMPLATES: Dict[str, str] = {
    # https://github.com/Redocly/redoc
    "redoc": """
<!DOCTYPE html>
<html>
    <head>
        <title>ReDoc</title>
        <!-- needed for adaptive design -->
        <meta charset="utf-8"/>
        <meta name="viewport" content="width=device-width, initial-scale=1">

        <!--
        ReDoc doesn't change outer page styles
        -->
        <style>
        body {{
            margin: 0;
            padding: 0;
        }}
        </style>
    </head>
    <body>
        <redoc spec-url='{spec_url}'></redoc>
        <script src="https://unpkg.com/redoc@v2.5.1/bundles/redoc.standalone.js"> </script>
    </body>
</html>""",
    # https://swagger.io
    "swagger": """
<!-- HTML for static distribution bundle build -->
<!DOCTYPE html>
<html lang="en">
    <head>
        <meta charset="UTF-8">
        <meta name="viewport" content="width=device-width, initial-scale=1" />
        <meta name="description" content="SwaggerUI"/>
        <title>SwaggerUI</title>
        <link rel="stylesheet" href="https://unpkg.com/swagger-ui-dist@5.11.0/swagger-ui.css" />
        <link rel="icon" href="data:image/svg+xml,<svg xmlns=%22http://www.w3.org/2000/svg%22 viewBox=%220 0 100 100%22><text y=%221em%22 font-size=%2280%22>📄</text></svg>">
    </head>

    <body>
        <div id="swagger-ui"></div>

        <script src="https://unpkg.com/swagger-ui-dist@5.11.0/swagger-ui-bundle.js" crossorigin></script>
        <script src="https://unpkg.com/swagger-ui-dist@5.11.0/swagger-ui-standalone-preset.js" crossorigin></script>
        <script>
        window.onload = function() {{
        var full = location.protocol + '//' + location.hostname + (location.port ? ':' + location.port : '');
        // Begin Swagger UI call region
        const ui = SwaggerUIBundle({{
            url: "{spec_url}",
            dom_id: '#swagger-ui',
            presets: [
                SwaggerUIBundle.presets.apis,
                SwaggerUIStandalonePreset
            ],
            layout: "StandaloneLayout",
            oauth2RedirectUrl: full + "/{spec_path}/swagger/oauth2-redirect.html",
        }});
        ui.initOAuth({{
            clientId: "{client_id}",
            clientSecret: "{client_secret}",
            realm: "{realm}",
            appName: "{app_name}",
            scopeSeparator: "{scope_separator}",
            additionalQueryStringParams: {additional_query_string_params},
            useBasicAuthenticationWithAccessCodeGrant: {use_basic_authentication_with_access_code_grant},
            usePkceWithAuthorizationCodeGrant: {use_pkce_with_authorization_code_grant},
        }});
        // End Swagger UI call region

        window.ui = ui;
        }}
    </script>
    </body>
</html>""",
    "swagger/oauth2-redirect.html": """
<!DOCTYPE html>
<html lang="en-US">
<head>
    <title>Swagger UI: OAuth2 Redirect</title>
</head>
<body>
<script>
    'use strict';
    function run () {{
        var oauth2 = window.opener.swaggerUIRedirectOauth2;
        var sentState = oauth2.state;
        var redirectUrl = oauth2.redirectUrl;
        var isValid, qp, arr;

        if (/code|token|error/.test(window.location.hash)) {{
            qp = window.location.hash.substring(1);
        }} else {{
            qp = location.search.substring(1);
        }}

        arr = qp.split("&");
        arr.forEach(function (v,i,_arr) {{ _arr[i] = '"' + v.replace('=', '":"') + '"';}});
        qp = qp ? JSON.parse('{{' + arr.join() + '}}',
                function (key, value) {{
                    return key === "" ? value : decodeURIComponent(value);
                }}
        ) : {{}};

        isValid = qp.state === sentState;

        if ((
          oauth2.auth.schema.get("flow") === "accessCode" ||
          oauth2.auth.schema.get("flow") === "authorizationCode" ||
          oauth2.auth.schema.get("flow") === "authorization_code"
        ) && !oauth2.auth.code) {{
            if (!isValid) {{
                oauth2.errCb({{
                    authId: oauth2.auth.name,
                    source: "auth",
                    level: "warning",
                    message: "Authorization may be unsafe, passed state was changed in server Passed state wasn't returned from auth server"
                }});
            }}

            if (qp.code) {{
                delete oauth2.state;
                oauth2.auth.code = qp.code;
                oauth2.callback({{auth: oauth2.auth, redirectUrl: redirectUrl}});
            }} else {{
                let oauthErrorMsg;
                if (qp.error) {{
                    oauthErrorMsg = "["+qp.error+"]: " +
                        (qp.error_description ? qp.error_description+ ". " : "no accessCode received from the server. ") +
                        (qp.error_uri ? "More info: "+qp.error_uri : "");
                }}

                oauth2.errCb({{
                    authId: oauth2.auth.name,
                    source: "auth",
                    level: "error",
                    message: oauthErrorMsg || "[Authorization failed]: no accessCode received from the server"
                }});
            }}
        }} else {{
            oauth2.callback({{auth: oauth2.auth, token: qp, isValid: isValid, redirectUrl: redirectUrl}});
        }}
        window.close();
    }}

    window.addEventListener('DOMContentLoaded', function () {{
      run();
    }});
</script>
</body>
</html>""",
    "scalar": """
<!doctype html>
<html>
  <head>
    <title>API Reference</title>
    <meta charset="utf-8" />
    <meta
      name="viewport"
      content="width=device-width, initial-scale=1" />
    <style>
      body {{
        margin: 0;
      }}
    </style>
    <link rel="icon" href="data:image/svg+xml,<svg xmlns=%22http://www.w3.org/2000/svg%22 viewBox=%220 0 100 100%22><text y=%221em%22 font-size=%2280%22>📄</text></svg>">
  </head>
  <body>
    <script
      id="api-reference"
      data-url="{spec_url}">
    </script>
    <script>
      var configuration = {{
        theme: 'purple',
      }}

      var apiReference = document.getElementById('api-reference')
      apiReference.dataset.configuration = JSON.stringify(configuration)
    </script>
    <script src="https://cdn.jsdelivr.net/npm/@scalar/api-reference"></script>
  </body>
</html>""",
}

try:
    from offapi import OpenAPITemplate

    PAGE_TEMPLATES = {
        "redoc": OpenAPITemplate.REDOC.value,
        "swagger": OpenAPITemplate.SWAGGER.value,
        "scalar": OpenAPITemplate.SCALAR.value,
    }
except ImportError:
    PAGE_TEMPLATES = ONLINE_PAGE_TEMPLATES



from typing import (
    Any,
    Callable,
    Dict,
    Iterator,
    List,
    Mapping,
    Optional,
    Protocol,
    Sequence,
    Type,
    TypeVar,
    Union,
)

from spectree._pydantic import BaseModel

BaseModelSubclassType = TypeVar("BaseModelSubclassType", bound=BaseModel)
ModelType = Type[BaseModelSubclassType]
OptionalModelType = Optional[ModelType]
NamingStrategy = Callable[[ModelType], str]
NestedNamingStrategy = Callable[[str, str], str]


class MultiDict(Protocol):
    def get(self, key: str) -> Optional[str]:
        pass

    def getlist(self, key: str) -> List[str]:
        pass

    def __iter__(self) -> Iterator[str]:
        pass


class MultiDictStarlette(Protocol):
    def __iter__(self) -> Iterator[str]:
        pass

    def getlist(self, key: Any) -> List[Any]:
        pass

    def __getitem__(self, key: Any) -> Any:
        pass


class FunctionDecorator(Protocol):
    resp: Any
    tags: Sequence[Any]
    security: Union[None, Dict, List[Any]]
    deprecated: bool
    path_parameter_descriptions: Optional[Mapping[str, str]]
    _decorator: Any


JsonType = Union[None, int, str, bool, List["JsonType"], Dict[str, "JsonType"]]



import re
from typing import Any, Iterator, List, Mapping, Optional, Tuple, Union

from werkzeug.datastructures import Headers
from werkzeug.routing import parse_converter_args

from spectree.plugins.base import BasePlugin
from spectree.utils import get_multidict_items

RE_FLASK_RULE = re.compile(
    r"""
    (?P<static>[^<]*)                           # static rule data
    <
    (?:
        (?P<converter>[a-zA-Z_][a-zA-Z0-9_]*)   # converter name
        (?:\((?P<args>.*?)\))?                  # converter arguments
        \:                                      # variable delimiter
    )?
    (?P<variable>[a-zA-Z_][a-zA-Z0-9_]*)        # variable name
    >
    """,
    re.VERBOSE,
)


def werkzeug_parse_rule(
    rule: str,
) -> Iterator[Tuple[Optional[str], Optional[str], str]]:
    """A copy of werkzeug.parse_rule which is now removed.

    Parse a rule and return it as generator. Each iteration yields tuples
    in the form ``(converter, arguments, variable)``. If the converter is
    `None` it's a static url part, otherwise it's a dynamic one.
    """
    pos = 0
    end = len(rule)
    do_match = RE_FLASK_RULE.match
    used_names = set()
    while pos < end:
        m = do_match(rule, pos)
        if m is None:
            break
        data = m.groupdict()
        if data["static"]:
            yield None, None, data["static"]
        variable = data["variable"]
        converter = data["converter"] or "default"
        if variable in used_names:
            raise ValueError(f"variable name {variable!r} used twice.")
        used_names.add(variable)
        yield converter, data["args"] or None, variable
        pos = m.end()
    if pos < end:
        remaining = rule[pos:]
        if ">" in remaining or "<" in remaining:
            raise ValueError(f"malformed url rule: {rule!r}")
        yield None, None, remaining


def flask_response_unpack(
    resp: Any,
) -> Tuple[Any, int, Union[List[Tuple[str, str]], Headers]]:
    """Parse Flask response object into a tuple of (payload, status_code, headers)."""
    status = 200
    headers: List[Tuple[str, str]] = []
    payload = None
    if not isinstance(resp, tuple):
        return resp, status, headers
    if len(resp) == 1:
        payload = resp[0]
    elif len(resp) == 2:
        payload = resp[0]
        if isinstance(resp[1], int):
            status = resp[1]
        else:
            headers = resp[1]
    elif len(resp) == 3:
        payload, status, headers = resp
    else:
        raise ValueError(
            f"Invalid return tuple: {resp}, expect (body,), (body, status), "
            "(body, headers), or (body, status, headers)."
        )
    return payload, status, headers


class WerkzeugPlugin(BasePlugin):
    blueprint_state = None

    def get_current_app(self):
        raise NotImplementedError()

    def is_app_response(self, resp) -> bool:
        raise NotImplementedError()

    def make_response_with_addition(self, *args):
        """This method is derived from Flask's `make_response` method."""
        current_app = self.get_current_app()
        if len(args) == 1:
            args = args[0]
        return current_app.make_response(args)

    @staticmethod
    def is_blueprint(app) -> bool:
        raise NotImplementedError()

    def find_routes(self):
        # https://werkzeug.palletsprojects.com/en/stable/routing/#werkzeug.routing.Rule
        for rule in self.get_current_app().url_map.iter_rules():
            if any(
                str(rule).startswith(path)
                for path in (f"/{self.config.path}", "/static")
            ):
                continue
            if rule.endpoint.startswith("openapi"):
                continue
            if getattr(rule, "websocket", False):
                continue
            if (
                self.blueprint_state
                and self.blueprint_state.url_prefix
                and (
                    not str(rule).startswith(self.blueprint_state.url_prefix)
                    or str(rule).startswith(
                        "/".join([self.blueprint_state.url_prefix, self.config.path])
                    )
                )
            ):
                continue
            yield rule

    def bypass(self, func, method):
        return method in ["HEAD", "OPTIONS"]

    def parse_func(self, route: Any):
        if self.blueprint_state:
            func = self.blueprint_state.app.view_functions[route.endpoint]
        else:
            func = self.get_current_app().view_functions[route.endpoint]

        # view class: https://flask.palletsprojects.com/en/1.1.x/views/
        view_cls = getattr(func, "view_class", None)
        if view_cls:
            for method in route.methods:
                view = getattr(view_cls, method.lower(), None)
                if view:
                    yield method, view
        else:
            for method in route.methods:
                yield method, func

    def parse_path(
        self,
        route: Optional[Mapping[str, str]],
        path_parameter_descriptions: Optional[Mapping[str, str]],
    ) -> Tuple[str, list]:
        subs = []
        parameters = []

        for converter, arguments, variable in werkzeug_parse_rule(str(route)):
            if converter is None:
                subs.append(variable)
                continue
            subs.append(f"{{{variable}}}")

            args: tuple = ()
            kwargs: dict = {}

            if arguments:
                args, kwargs = parse_converter_args(arguments)

            schema = None
            if converter == "any":
                schema = {
                    "type": "string",
                    "enum": args,
                }
            elif converter == "int":
                schema = {
                    "type": "integer",
                    "format": "int32",
                }
                if "max" in kwargs:
                    schema["maximum"] = kwargs["max"]
                if "min" in kwargs:
                    schema["minimum"] = kwargs["min"]
            elif converter == "float":
                schema = {
                    "type": "number",
                    "format": "float",
                }
            elif converter == "uuid":
                schema = {
                    "type": "string",
                    "format": "uuid",
                }
            elif converter == "path":
                schema = {
                    "type": "string",
                    "format": "path",
                }
            elif converter == "string":
                schema = {
                    "type": "string",
                }
                for prop in ["length", "maxLength", "minLength"]:
                    if prop in kwargs:
                        schema[prop] = kwargs[prop]
            elif converter == "default":
                schema = {"type": "string"}

            description = (
                path_parameter_descriptions.get(variable, "")
                if path_parameter_descriptions
                else ""
            )
            parameters.append(
                {
                    "name": variable,
                    "in": "path",
                    "required": True,
                    "schema": schema,
                    "description": description,
                }
            )

        return "".join(subs), parameters

    def fill_form(self, request) -> dict:
        req_data = get_multidict_items(request.form)
        req_data.update(get_multidict_items(request.files) if request.files else {})
        return req_data

    def register_route(self, app):
        app.add_url_rule(
            rule=self.config.spec_url,
            endpoint=f"openapi_{self.config.path}",
            view_func=lambda: self.get_current_app().json.response(self.spectree.spec),
        )

        if self.is_blueprint(app):

            def gen_doc_page(ui):
                spec_url = self.config.spec_url
                if self.blueprint_state.url_prefix is not None:
                    spec_url = "/".join(
                        (
                            self.blueprint_state.url_prefix.rstrip("/"),
                            self.config.spec_url.lstrip("/"),
                        )
                    )

                return self.config.page_templates[ui].format(
                    spec_url=spec_url,
                    spec_path=self.config.path,
                    **self.config.swagger_oauth2_config(),
                )

            for ui in self.config.page_templates:
                app.add_url_rule(
                    rule=f"/{self.config.path}/{ui}/",
                    endpoint=f"openapi_{self.config.path}_{ui.replace('.', '_')}",
                    view_func=lambda ui=ui: gen_doc_page(ui),
                )

            app.record(lambda state: setattr(self, "blueprint_state", state))
        else:
            for ui in self.config.page_templates:
                app.add_url_rule(
                    rule=f"/{self.config.path}/{ui}/",
                    endpoint=f"openapi_{self.config.path}_{ui}",
                    view_func=lambda ui=ui: self.config.page_templates[ui].format(
                        spec_url=self.config.spec_url,
                        spec_path=self.config.path,
                        **self.config.swagger_oauth2_config(),
                    ),
                )



import string
from typing import Type

import pytest
from pydantic import ValidationError

from spectree import SecurityScheme
from spectree.config import Configuration

from .common import SECURITY_SCHEMAS, WRONG_SECURITY_SCHEMAS_DATA


def test_config_license():
    config = Configuration(license={"name": "MIT"})
    assert config.license.name == "MIT"

    config = Configuration(
        license={"name": "MIT", "url": "https://opensource.org/licenses/MIT"}
    )
    assert config.license.name == "MIT"
    assert str(config.license.url) == "https://opensource.org/licenses/MIT"

    with pytest.raises(ValidationError):
        Configuration(license={"name": "MIT", "url": "url"})


def test_config_contact():
    config = Configuration(contact={"name": "John"})
    assert config.contact.name == "John"

    config = Configuration(contact={"name": "John", "url": "https://example.com"})
    assert config.contact.name == "John"
    assert str(config.contact.url).rstrip("/") == "https://example.com"

    config = Configuration(contact={"name": "John", "email": "hello@github.com"})
    assert config.contact.name == "John"
    assert config.contact.email == "hello@github.com"

    with pytest.raises(ValidationError):
        Configuration(contact={"name": "John", "url": "url"})


@pytest.mark.parametrize(("secure_item"), SECURITY_SCHEMAS)
def test_update_security_scheme(secure_item: Type[SecurityScheme]):
    # update and validate each schema type
    config = Configuration(
        security_schemes=[SecurityScheme(name=secure_item.name, data=secure_item.data)]
    )
    assert config.security_schemes
    assert config.security_schemes[0].name == secure_item.name
    assert config.security_schemes[0].data == secure_item.data


def test_update_security_schemes():
    # update and validate ALL schemas types
    config = Configuration(security_schemes=SECURITY_SCHEMAS)
    assert config.security_schemes == SECURITY_SCHEMAS


@pytest.mark.parametrize(("secure_item"), SECURITY_SCHEMAS)
def test_update_security_scheme_wrong_type(secure_item: SecurityScheme):
    # update and validate each schema type
    with pytest.raises(ValidationError):
        secure_item.data.type += "_wrong"  # type: ignore


@pytest.mark.parametrize(
    "symbol", [symb for symb in string.punctuation if symb not in "-._"]
)
@pytest.mark.parametrize(("secure_item"), SECURITY_SCHEMAS)
def test_update_security_scheme_wrong_name(secure_item: SecurityScheme, symbol: str):
    # update and validate each schema name
    with pytest.raises(ValidationError):
        secure_item.name += symbol

    with pytest.raises(ValidationError):
        secure_item.name = symbol + secure_item.name


@pytest.mark.parametrize(("secure_item"), WRONG_SECURITY_SCHEMAS_DATA)
def test_update_security_scheme_wrong_data(secure_item: dict):
    # update and validate each schema type
    with pytest.raises(ValidationError):
        SecurityScheme(**secure_item)



from random import randint
from typing import List

import pytest
from flask import Flask, jsonify, make_response, request
from flask.views import MethodView

from spectree import Response, SpecTree

from .common import (
    JSON,
    Cookies,
    CustomError,
    Form,
    FormFileUpload,
    Headers,
    ListJSON,
    OptionalAliasResp,
    Order,
    Query,
    QueryList,
    Resp,
    RespFromAttrs,
    RespObject,
    RootResp,
    StrDict,
    UserXmlData,
    api_tag,
    get_root_resp_data,
)

# import tests to execute
from .flask_imports import *  # NOQA


def before_handler(req, resp, err, _):
    if err:
        resp.headers["X-Error"] = "Validation Error"


def after_handler(req, resp, err, _):
    resp.headers["X-Validation"] = "Pass"


def api_after_handler(req, resp, err, _):
    resp.headers["X-API"] = "OK"


api = SpecTree("flask", before=before_handler, after=after_handler, annotations=True)
app = Flask(__name__)
app.config["TESTING"] = True


class Ping(MethodView):
    @api.validate(
        headers=Headers, resp=Response(HTTP_202=StrDict), tags=["test", "health"]
    )
    def get(self):
        """summary

        description"""
        return jsonify(msg="pong"), 202, request.context.headers.model_dump()


class FileUploadView(MethodView):
    @api.validate(
        form=FormFileUpload,
    )
    def post(self, form: FormFileUpload):
        upload = form.file
        assert upload
        return {"content": upload.stream.read().decode("utf-8"), "other": form.other}


class User(MethodView):
    @api.validate(
        query=Query,
        json=JSON,
        form=Form,
        cookies=Cookies,
        resp=Response(HTTP_200=Resp, HTTP_401=None),
        tags=[api_tag, "test"],
        after=api_after_handler,
    )
    def post(self, name):
        data_src = request.context.json or request.context.form
        score = [randint(0, int(data_src.limit)) for _ in range(5)]
        score.sort(reverse=request.context.query.order)
        assert request.context.cookies.pub == "abcdefg"
        assert request.cookies["pub"] == "abcdefg"
        return jsonify(name=data_src.name, score=score)


class UserAnnotated(MethodView):
    @api.validate(
        resp=Response(HTTP_200=Resp, HTTP_401=None),
        tags=[api_tag, "test"],
        after=api_after_handler,
    )
    def post(self, name, query: Query, json: JSON, form: Form, cookies: Cookies):
        data_src = json or form
        score = [randint(0, int(data_src.limit)) for _ in range(5)]
        score.sort(reverse=query.order == Order.desc)
        assert cookies.pub == "abcdefg"
        assert request.cookies["pub"] == "abcdefg"
        return jsonify(name=data_src.name, score=score)


class UserSkip(MethodView):
    @api.validate(
        query=Query,
        json=JSON,
        cookies=Cookies,
        resp=Response(HTTP_200=Resp, HTTP_401=None),
        tags=[api_tag, "test"],
        after=api_after_handler,
        skip_validation=True,
    )
    def post(self, name):
        response_format = request.args.get("response_format")
        assert response_format in ("json", "xml")
        data_src = request.get_json() or request.get_data()
        score = [randint(0, int(data_src["limit"])) for _ in range(5)]
        score.sort(reverse=(int(request.args["order"]) == Order.desc))
        assert request.cookies["pub"] == "abcdefg"
        if response_format == "json":
            return jsonify(name=name, x_score=score)
        else:
            return app.response_class(
                UserXmlData(name=name, score=score).dump_xml(),
                content_type="text/xml",
            )


class UserModel(MethodView):
    @api.validate(
        query=Query,
        json=JSON,
        cookies=Cookies,
        resp=Response(HTTP_200=Resp, HTTP_401=None),
        tags=[api_tag, "test"],
        after=api_after_handler,
    )
    def post(self, name, query: Query, json: JSON, form: Form, cookies: Cookies):
        data_src = json or form
        score = [randint(0, int(data_src.limit)) for _ in range(5)]
        score.sort(reverse=(query.order == Order.desc))
        assert cookies.pub == "abcdefg"
        assert request.cookies["pub"] == "abcdefg"
        return Resp(name=data_src.name, score=score)


class UserAddress(MethodView):
    @api.validate(
        query=Query,
        path_parameter_descriptions={
            "name": "The name that uniquely identifies the user.",
            "non-existent-param": "description",
        },
    )
    def get(self, name, address_id):
        return None


class NoResponseView(MethodView):
    @api.validate(
        resp=Response(HTTP_200=None),  # response is None
    )
    def get(self):
        return {}

    @api.validate(
        json=StrDict,  # resp is missing completely
    )
    def post(self, json: JSON):
        return {}


class SetCookiesView(MethodView):
    @api.validate(
        resp=Response(HTTP_200=StrDict),
    )
    def get(self):
        # related to GitHub issue #415
        resp = make_response(jsonify(msg="ping"))
        resp.set_cookie("foo", "hello")
        resp.set_cookie("bar", "world")
        return resp


class ListJsonView(MethodView):
    @api.validate(json=ListJSON)
    def post(self):
        return {}


class QueryWithList(MethodView):
    @api.validate(query=QueryList)
    def get(self):
        assert request.context.query.ids == [1, 2, 3]
        return {}


class ReturnListView(MethodView):
    @api.validate(
        resp=Response(HTTP_200=List[JSON]),
    )
    def get(self):
        pre_serialize = bool(int(request.args.get("pre_serialize", default=0)))
        data = [JSON(name="user1", limit=1), JSON(name="user2", limit=2)]
        return [entry.model_dump() if pre_serialize else entry for entry in data]


class ReturnMakeResponseView(MethodView):
    @api.validate(
        json=JSON,
        headers=Headers,
        resp=Response(HTTP_201=Resp),
    )
    def post(self):
        model_data = request.context.json
        headers = request.context.headers
        response = make_response(
            Resp(name=model_data.name, score=[model_data.limit]).model_dump(),
            201,
            headers,
        )
        response.set_cookie(
            key="test_cookie",
            value=model_data.name,
            secure=True,
            httponly=True,
            samesite="Strict",
        )
        return response

    @api.validate(
        query=JSON,
        headers=Headers,
        resp=Response(HTTP_201=Resp),
    )
    def get(self):
        model_data = request.context.query
        headers = request.context.headers
        response = make_response(
            Resp(name=model_data.name, score=[model_data.limit]).model_dump(),
            201,
            headers,
        )
        response.set_cookie(
            key="test_cookie",
            value=model_data.name,
            secure=True,
            httponly=True,
            samesite="Strict",
        )
        return response


class ReturnRootView(MethodView):
    @api.validate(resp=Response(HTTP_200=RootResp))
    def get(self):
        return get_root_resp_data(
            pre_serialize=bool(int(request.args.get("pre_serialize", default=0))),
            return_what=request.args.get("return_what", default="RootResp"),
        )


class ReturnModelView(MethodView):
    @api.validate()
    def get(self):
        return get_root_resp_data(
            pre_serialize=False,
            return_what=request.args.get("return_what", default="RootResp"),
        )


class ReturnOptionalAlias(MethodView):
    @api.validate(resp=Response(HTTP_200=OptionalAliasResp))
    def get(self):
        return {"schema": "test"}


class CustomErrorView(MethodView):
    @api.validate(resp=Response(HTTP_200=CustomError))
    def post(self, json: CustomError):
        return jsonify(foo="bar")


class StringStatusView(MethodView):
    @api.validate()
    def get(self):
        return "Response text string", 200


class ForcedSerializeView(MethodView):
    @api.validate(resp=Response(HTTP_200=RespFromAttrs), force_resp_serialize=True)
    def get(self):
        return RespObject(name="flask", score=[1, 2, 3], comment="hello")


app.add_url_rule("/ping", view_func=Ping.as_view("ping"))
app.add_url_rule("/api/user/<name>", view_func=User.as_view("user"), methods=["POST"])
app.add_url_rule(
    "/api/user_annotated/<name>",
    view_func=UserAnnotated.as_view("user_annotated"),
    methods=["POST"],
)
app.add_url_rule(
    "/api/user_skip/<name>",
    view_func=UserSkip.as_view("user_skip"),
    methods=["POST"],
)
app.add_url_rule(
    "/api/user_model/<name>",
    view_func=UserModel.as_view("user_model"),
    methods=["POST"],
)
app.add_url_rule(
    "/api/user/<name>/address/<address_id>",
    view_func=UserAddress.as_view("user_address"),
    methods=["GET"],
)
app.add_url_rule(
    "/api/no_response",
    view_func=NoResponseView.as_view("no_response_view"),
)
app.add_url_rule(
    "/api/file_upload",
    view_func=FileUploadView.as_view("file_upload_view"),
)
app.add_url_rule(
    "/api/list_json",
    view_func=ListJsonView.as_view("list_json_view"),
)
app.add_url_rule(
    "/api/query_list",
    view_func=QueryWithList.as_view("query_list_view"),
)
app.add_url_rule(
    "/api/return_list",
    view_func=ReturnListView.as_view("return_list_view"),
)
app.add_url_rule(
    "/api/return_make_response",
    view_func=ReturnMakeResponseView.as_view("return_make_response"),
)
app.add_url_rule(
    "/api/return_optional_alias",
    view_func=ReturnOptionalAlias.as_view("return_optional_alias"),
)
app.add_url_rule(
    "/api/return_root",
    view_func=ReturnRootView.as_view("return_root_view"),
)
app.add_url_rule(
    "/api/return_model",
    view_func=ReturnModelView.as_view("return_model_view"),
)
app.add_url_rule(
    "/api/return_string_status",
    view_func=StringStatusView.as_view("string_status_view"),
)
app.add_url_rule(
    "/api/custom_error",
    view_func=CustomErrorView.as_view("custom_error_view"),
)
app.add_url_rule(
    "/api/force_serialize",
    view_func=ForcedSerializeView.as_view("force_serialize_view"),
)

# INFO: ensures that spec is calculated and cached _after_ registering
# view functions for validations. This enables tests to access `api.spec`
# without app_context.
with app.app_context():
    _ = api.spec


api.register(app)


@pytest.fixture
def client():
    with app.test_client() as client:
        yield client


@pytest.fixture
def test_client_and_api(request):
    api_args = ["flask"]
    api_kwargs = {}
    endpoint_kwargs = {
        "headers": Headers,
        "resp": Response(HTTP_200=StrDict),
        "tags": ["test", "health"],
    }
    if hasattr(request, "param"):
        api_args.extend(request.param.get("api_args", ()))
        api_kwargs.update(request.param.get("api_kwargs", {}))
        endpoint_kwargs.update(request.param.get("endpoint_kwargs", {}))

    api = SpecTree(*api_args, **api_kwargs)
    app = Flask(__name__)
    app.config["TESTING"] = True

    class Ping(MethodView):
        @api.validate(**endpoint_kwargs)
        def get(self):
            """summary

            description"""
            return jsonify(msg="pong")

    app.add_url_rule("/ping", view_func=Ping.as_view("ping"))

    # INFO: ensures that spec is calculated and cached _after_ registering
    # view functions for validations. This enables tests to access `api.spec`
    # without app_context.
    with app.app_context():
        _ = api.spec
    api.register(app)

    with app.test_client() as test_client:
        yield test_client, api



import json

import pytest
from pydantic import BaseModel, computed_field

from spectree.models import ValidationError
from spectree.response import DEFAULT_CODE_DESC, Response
from spectree.spec import SpecTree
from spectree.utils import (
    get_model_schema,
    has_model,
    json_compatible_deepcopy,
    parse_code,
    parse_comments,
    parse_name,
    parse_params,
    parse_request,
    parse_resp,
)

from .common import DefaultEnumValue, DemoModel, DemoQuery, Numeric, get_model_path_key

api = SpecTree()


def undecorated_func():
    """summary

    description"""


@api.validate(json=DemoModel, resp=Response(HTTP_200=DemoModel))
def demo_func():
    """
    summary

    description"""


@api.validate(query=DemoQuery)
def demo_func_with_query():
    """
    a summary

    a description
    """


class DemoClass:
    @api.validate(query=DemoModel)
    def demo_method(self):
        """summary

        description
        """


demo_class = DemoClass()


@pytest.mark.parametrize(
    "docstring, expected_summary, expected_description",
    [
        pytest.param(None, None, None, id="no-docstring"),
        pytest.param("", "", None, id="empty-docstring"),
        pytest.param("   ", "", None, id="all-whitespace-docstring"),
        pytest.param("summary", "summary", None, id="single-line-docstring"),
        pytest.param(
            "   summary   ", "summary", None, id="single-line-docstring-with-whitespace"
        ),
        pytest.param(
            "summary first line\nsummary second line",
            "summary first line summary second line",
            None,
            id="multi-line-docstring-without-empty-line",
        ),
        pytest.param(
            "  summary first line \n summary second line  ",
            "summary first line  summary second line",
            None,
            id="multi-line-docstring-without-empty-line-whitespace",
        ),
        pytest.param(
            "summary\n\ndescription",
            "summary",
            "description",
            id="multi-line-docstring-with-empty-line",
        ),
        pytest.param(
            "   summary   \n\n   description  ",
            "summary",
            "description",
            id="multi-line-docstring-with-empty-line-whitespace",
        ),
        pytest.param(
            "summary\n\t   \ndescription",
            "summary",
            "description",
            id="multi-line-docstring-with-whitespace-line",
        ),
        pytest.param(
            "summary\n  \n  \n  \n  \n  \ndescription",
            "summary",
            "description",
            id="multi-line-docstring-with-multiple-whitespace-lines",
        ),
        pytest.param(
            "summary first line\nsummary second line\nsummary third line"
            "\n\t   \n"
            "description first line\ndescription second line\ndescription third line",
            "summary first line summary second line summary third line",
            "description first line description second line description third line",
            id="large-multi-line-docstring-with-whitespace-line",
        ),
        pytest.param(
            "summary first line\nsummary second line\ftruncated part",
            "summary first line summary second line",
            None,
            id="multi-line-docstring-without-empty-line-and-truncation-char",
        ),
        pytest.param(
            "summary first line\nsummary second line\nsummary third line"
            "\n\t   \n"
            "description first line\ndescription second line\ndescription third line"
            "\ftruncated part",
            "summary first line summary second line summary third line",
            "description first line description second line description third line",
            id="large-multi-line-docstring-with-whitespace-line-and-truncation-char",
        ),
        pytest.param(
            "summary first line\nsummary second line\n"
            "\t   \n"
            "description first line   \ndescription second line\n"
            "\t   \n"
            "description second paragraph   \n"
            "\n \n \n"
            "description third paragraph\ndescription third paragraph second line",
            "summary first line summary second line",
            "description first line    description second line"
            "\n\n"
            "description second paragraph"
            "\n\n"
            "description third paragraph description third paragraph second line",
            id="large-multi-line-docstring-with-multiple-paragraphs",
        ),
        pytest.param(
            "\tcode block while indented\n"
            "\t\n"
            "\tdescription first paragraph\n"
            "\t\n"
            "\t\tcode block\n"
            "\t\n"
            "\tdescription third paragraph\n",
            "code block while indented",
            "description first paragraph"
            "\n\n"
            "        code block"
            "\n\n"
            "description third paragraph",
            id="multi-line-docstring-with-code-block",
        ),
    ],
)
def test_parse_comments(docstring, expected_summary, expected_description):
    def func():
        pass

    func.__doc__ = docstring

    assert parse_comments(func) == (expected_summary, expected_description)


@pytest.mark.parametrize(
    "func, expected_summary, expected_description",
    [
        pytest.param(lambda x: x, None, None, id="lambda"),
        pytest.param(
            undecorated_func, "summary", "description", id="undecorated-function"
        ),
        pytest.param(demo_func, "summary", "description", id="decorated-function"),
        pytest.param(
            demo_class.demo_method, "summary", "description", id="class-method"
        ),
    ],
)
def test_parse_comments_with_different_callable_types(
    func, expected_summary, expected_description
):
    assert parse_comments(func) == (expected_summary, expected_description)


def test_parse_code():
    with pytest.raises(TypeError):
        assert parse_code(200) == 200

    assert parse_code("200") == ""
    assert parse_code("HTTP_404") == "404"


def test_parse_name():
    assert parse_name(lambda x: x) == "<lambda>"
    assert parse_name(undecorated_func) == "undecorated_func"
    assert parse_name(demo_func) == "demo_func"
    assert parse_name(demo_class.demo_method) == "demo_method"


def test_has_model():
    assert not has_model(undecorated_func)
    assert has_model(demo_func)
    assert has_model(demo_class.demo_method)


def test_parse_resp():
    assert parse_resp(undecorated_func) == {}
    resp_spec = parse_resp(demo_func)

    assert resp_spec["422"]["description"] == DEFAULT_CODE_DESC["HTTP_422"]
    model_path_key = get_model_path_key(
        f"{ValidationError.__module__}.{ValidationError.__name__}"
    )
    assert (
        resp_spec["422"]["content"]["application/json"]["schema"]["$ref"]
        == f"#/components/schemas/{model_path_key}"
    )
    model_path_key = get_model_path_key(f"{DemoModel.__module__}.{DemoModel.__name__}")
    assert (
        resp_spec["200"]["content"]["application/json"]["schema"]["$ref"]
        == f"#/components/schemas/{model_path_key}"
    )


def test_parse_request():
    model_path_key = get_model_path_key(f"{DemoModel.__module__}.{DemoModel.__name__}")
    assert (
        parse_request(demo_func)["content"]["application/json"]["schema"]["$ref"]
        == f"#/components/schemas/{model_path_key}"
    )
    assert parse_request(demo_class.demo_method) == {}


def test_parse_params():
    models = {
        get_model_path_key(
            f"{DemoModel.__module__}.{DemoModel.__name__}"
        ): DemoModel.model_json_schema(ref_template="#/components/schemas/{model}")
    }
    assert parse_params(demo_func, [], models) == []
    params = parse_params(demo_class.demo_method, [], models)
    assert len(params) == 3
    assert params[0] == {
        "name": "uid",
        "in": "query",
        "required": True,
        "description": "",
        "schema": {"title": "Uid", "type": "integer"},
    }
    assert params[2]["description"] == "user name"


def test_parse_params_with_route_param_keywords():
    models = {
        get_model_path_key("tests.common.DemoQuery"): DemoQuery.model_json_schema(
            ref_template="#/components/schemas/{model}"
        )
    }
    params = parse_params(demo_func_with_query, [], models)
    assert params == [
        {
            "name": "names1",
            "in": "query",
            "required": True,
            "description": "",
            "schema": {"title": "Names1", "type": "array", "items": {"type": "string"}},
        },
        {
            "name": "names2",
            "in": "query",
            "required": True,
            "description": "",
            "schema": {
                "title": "Names2",
                "type": "array",
                "items": {"type": "string"},
                "non_keyword": "dummy",
            },
            "style": "matrix",
            "explode": True,
        },
    ]


def test_json_compatible_schema():
    schema = get_model_schema(Numeric)

    with pytest.raises(ValueError):
        json.dumps(schema, allow_nan=False)

    json_schema = json_compatible_deepcopy(schema)
    assert json.dumps(json_schema, allow_nan=False)

    schema = get_model_schema(DefaultEnumValue)
    json_schema = json_compatible_deepcopy(schema)


def test_get_model_schema_mode_parameter():
    """Test get_model_schema mode parameter for Pydantic v2"""

    class TestModel(BaseModel):
        """Model with computed field"""

        name: str
        value: int

        @computed_field
        @property
        def computed_name(self) -> str:
            """Computed field - only in serialization"""
            return f"computed_{self.name}"

    # Test validation mode - computed fields excluded
    validation_schema = get_model_schema(TestModel, mode="validation")
    assert "name" in validation_schema["properties"]
    assert "value" in validation_schema["properties"]
    assert "computed_name" not in validation_schema["properties"], (
        "Computed field should NOT be in validation mode"
    )

    # Test serialization mode - computed fields included
    serialization_schema = get_model_schema(TestModel, mode="serialization")
    assert "name" in serialization_schema["properties"]
    assert "value" in serialization_schema["properties"]
    assert "computed_name" in serialization_schema["properties"], (
        "Computed field SHOULD be in serialization mode"
    )

    # Verify computed field is marked as readOnly and required
    assert serialization_schema["properties"]["computed_name"].get("readOnly") is True
    assert "computed_name" in serialization_schema["required"]



from typing import List, get_type_hints

import pytest
from pydantic import BaseModel

from spectree.models import ValidationError
from spectree.response import DEFAULT_CODE_DESC, Response
from spectree.utils import gen_list_model

from .common import JSON, DemoModel, get_model_path_key


class NormalClass:
    pass


def test_init_response():
    for args, kwargs in [
        ([200], {}),
        (["HTTP_110"], {}),
        ([], {"HTTP_200": NormalClass}),
        ([], {"HTTP_200": (NormalClass, "custom code description")}),
        ([], {"HTTP_200": (DemoModel, 1)}),
        ([], {"HTTP_200": (DemoModel,)}),
    ]:
        with pytest.raises(AssertionError):
            Response(*args, **kwargs)

    resp = Response("HTTP_200", HTTP_201=DemoModel)
    assert resp.has_model()
    assert resp.find_model(201) == DemoModel
    assert resp.code_descriptions.get("HTTP_200") is None
    assert resp.code_descriptions.get("HTTP_201") is None
    assert DemoModel in resp.models

    resp = Response(
        HTTP_200=None,
        HTTP_400=List[JSON],
        HTTP_401=DemoModel,
        HTTP_402=(None, "custom code description"),
        HTTP_403=(DemoModel, "custom code description"),
    )
    expect_400_model = gen_list_model(JSON)
    assert resp.has_model()
    assert resp.find_model(200) is None
    assert type(resp.find_model(400)) is type(expect_400_model) and get_type_hints(
        resp.find_model(400)
    ) == get_type_hints(expect_400_model)
    assert resp.find_model(401) == DemoModel
    assert resp.find_model(402) is None
    assert resp.find_model(403) == DemoModel
    assert resp.code_descriptions.get("HTTP_200") is None
    assert resp.code_descriptions.get("HTTP_401") is None
    assert resp.code_descriptions.get("HTTP_402") == "custom code description"
    assert resp.code_descriptions.get("HTTP_403") == "custom code description"
    assert DemoModel in resp.models

    assert not Response().has_model()


def test_response_add_model():
    resp = Response()

    resp.add_model(201, DemoModel)

    assert resp.find_model(201) == DemoModel


@pytest.mark.parametrize(
    "replace, expected_model",
    [
        pytest.param(True, JSON, id="replace-existing-model"),
        pytest.param(False, DemoModel, id="keep-existing-model"),
    ],
)
def test_response_add_model_when_model_already_exists(replace, expected_model):
    resp = Response()

    resp.add_model(201, DemoModel)
    resp.add_model(201, JSON, replace=replace)

    assert resp.find_model(201) is expected_model


def test_response_spec():
    resp = Response(
        "HTTP_200",
        HTTP_201=DemoModel,
        HTTP_401=(DemoModel, "custom code description"),
        HTTP_402=(None, "custom code description"),
    )
    resp.add_model(422, ValidationError)
    spec = resp.generate_spec()
    assert spec["200"]["description"] == DEFAULT_CODE_DESC["HTTP_200"]
    assert spec["201"]["description"] == DEFAULT_CODE_DESC["HTTP_201"]
    assert spec["422"]["description"] == DEFAULT_CODE_DESC["HTTP_422"]
    assert spec["401"]["description"] == "custom code description"
    assert spec["402"]["description"] == "custom code description"
    assert spec["201"]["content"]["application/json"]["schema"]["$ref"].split("/")[
        -1
    ] == get_model_path_key(f"{DemoModel.__module__}.{DemoModel.__name__}")
    assert spec["422"]["content"]["application/json"]["schema"]["$ref"].split("/")[
        -1
    ] == get_model_path_key(f"{ValidationError.__module__}.{ValidationError.__name__}")

    assert spec.get(200) is None
    assert spec.get(404) is None


def test_list_model():
    resp = Response(HTTP_200=List[JSON])
    model = resp.find_model(200)
    expect_model = gen_list_model(JSON)
    assert resp.expect_list_result(200)
    assert not resp.expect_list_result(500)
    assert get_type_hints(model) == get_type_hints(expect_model)
    assert type(model) is type(expect_model)
    assert issubclass(model, BaseModel)
    data = [
        {"name": "a", "limit": 1},
        {"name": "b", "limit": 2},
    ]
    instance = model.model_validate(data)
    items = instance.model_dump()
    if isinstance(items, dict):
        items = items["__root__"]
    for i, item in enumerate(items):
        obj = JSON.model_validate(item)
        assert obj.name == data[i]["name"]
        assert obj.limit == data[i]["limit"]



import json
import uuid
from contextlib import nullcontext as does_not_raise
from dataclasses import dataclass
from datetime import datetime
from typing import Any, Union

import pytest
from pydantic import ValidationError

from spectree._pydantic import SerializedPydanticResponse
from spectree._types import OptionalModelType
from spectree.plugins.base import (
    RawResponsePayload,
    ResponseValidationResult,
    validate_response,
)
from spectree.utils import gen_list_model
from tests.common import JSON, ComplexResp, Resp, RootResp, StrDict

RespList = gen_list_model(Resp)


@dataclass(frozen=True)
class DummyResponse:
    payload: bytes
    content_type: str


@pytest.mark.parametrize(
    [
        "validation_model",
        "response_payload",
        "expected_result",
    ],
    [
        (
            Resp,
            Resp(name="user1", score=[1, 2]),
            ResponseValidationResult({"name": "user1", "score": [1, 2]}),
        ),
        (
            Resp,
            {"name": "user1", "score": [1, 2]},
            ResponseValidationResult({"name": "user1", "score": [1, 2]}),
        ),
        (
            Resp,
            RawResponsePayload({"name": "user1", "score": [1, 2]}),
            ResponseValidationResult({"name": "user1", "score": [1, 2]}),
        ),
        (
            Resp,
            {},
            ValidationError,
        ),
        (
            Resp,
            {"name": "user1"},
            ValidationError,
        ),
        (
            RootResp,
            [1, 2, 3],
            ResponseValidationResult([1, 2, 3]),
        ),
        (
            RootResp,
            RawResponsePayload([1, 2, 3]),
            ResponseValidationResult([1, 2, 3]),
        ),
        (
            StrDict,
            StrDict.model_validate({"key1": "value1", "key2": "value2"}),
            ResponseValidationResult({"key1": "value1", "key2": "value2"}),
        ),
        (
            RootResp,
            {"name": "user2", "limit": 1},
            ResponseValidationResult({"name": "user2", "limit": 1}),
        ),
        (
            RootResp,
            RawResponsePayload({"name": "user2", "limit": 1}),
            ResponseValidationResult({"name": "user2", "limit": 1}),
        ),
        (
            RootResp,
            JSON(name="user3", limit=5),
            ResponseValidationResult({"name": "user3", "limit": 5}),
        ),
        (
            RootResp,
            RootResp.model_validate(JSON(name="user4", limit=23)),
            ResponseValidationResult({"name": "user4", "limit": 23}),
        ),
        (
            RootResp,
            {},
            ValidationError,
        ),
        (
            RespList,
            [],
            ResponseValidationResult([]),
        ),
        (
            RespList,
            [{"name": "user5", "score": [5, 10]}],
            ResponseValidationResult([{"name": "user5", "score": [5, 10]}]),
        ),
        (
            RespList,
            [Resp(name="user6", score=[10, 20]), Resp(name="user7", score=[30, 40])],
            ResponseValidationResult(
                [
                    {"name": "user6", "score": [10, 20]},
                    {"name": "user7", "score": [30, 40]},
                ]
            ),
        ),
        (
            None,
            {"user_id": "user1", "locale": "en-gb"},
            ResponseValidationResult({"user_id": "user1", "locale": "en-gb"}),
        ),
        (
            None,
            DummyResponse(payload="<html></html>".encode(), content_type="text/html"),
            ResponseValidationResult(
                DummyResponse(
                    payload="<html></html>".encode(), content_type="text/html"
                )
            ),
        ),
        (
            ComplexResp,
            ComplexResp(
                date=datetime(2025, 1, 1),
                uuid=uuid.UUID("48b417cd-a884-4e54-9f5b-85c584e5ce77"),
            ),
            ResponseValidationResult(
                {
                    "date": "2025-01-01T00:00:00",
                    "uuid": "48b417cd-a884-4e54-9f5b-85c584e5ce77",
                }
            ),
        ),
    ],
)
def test_validate_response(
    validation_model: OptionalModelType,
    response_payload: Any,
    expected_result: Union[ResponseValidationResult, ValidationError],
):
    runtime_expectation = (
        pytest.raises(ValidationError)
        if expected_result == ValidationError
        else does_not_raise()
    )
    with runtime_expectation:
        result = validate_response(
            validation_model=validation_model,
            response_payload=response_payload,
        )
        assert isinstance(result, ResponseValidationResult)
        payload = (
            ResponseValidationResult(json.loads(result.payload.data))
            if isinstance(result.payload, SerializedPydanticResponse)
            else result
        )
        assert payload == expected_result



import re
from enum import Enum
from typing import Any, Dict, Optional, Sequence, Set

from pydantic import (
    BaseModel,
    ConfigDict,
    Field,
    RootModel,
    field_validator,
    model_validator,
)
from pydantic_core import core_schema

# OpenAPI names validation regexp
OpenAPI_NAME_RE = re.compile(r"^[A-Za-z0-9-._]+")

_EMPTY_SET: set[None] = set()


class ExternalDocs(BaseModel):
    description: str = ""
    url: str


class Tag(BaseModel):
    """OpenAPI tag object"""

    name: str
    description: str = ""
    externalDocs: Optional[ExternalDocs] = None

    def __str__(self):
        return self.name


class ValidationErrorElement(BaseModel):
    """Model of a validation error response element."""

    loc: Sequence[str] = Field(
        ...,
        title="Missing field name",
    )
    msg: str = Field(
        ...,
        title="Error message",
    )
    type: str = Field(
        ...,
        title="Error type",
    )
    ctx: Optional[Dict[str, Any]] = Field(
        None,
        title="Error context",
    )


class ValidationError(RootModel[Sequence[ValidationErrorElement]]):
    """Model of a validation error response."""


class SecureType(str, Enum):
    HTTP = "http"
    API_KEY = "apiKey"
    OAUTH_TWO = "oauth2"
    OPEN_ID_CONNECT = "openIdConnect"


class InType(str, Enum):
    HEADER = "header"
    QUERY = "query"
    COOKIE = "cookie"


type_req_fields: Dict[SecureType, Set[str]] = {
    SecureType.HTTP: {"scheme"},
    SecureType.API_KEY: {"name", "in"},
    SecureType.OAUTH_TWO: {"flows"},
    SecureType.OPEN_ID_CONNECT: {"openIdConnectUrl"},
}


class SecuritySchemeData(BaseModel):
    """
    Security scheme data
    https://github.com/OAI/OpenAPI-Specification/blob/main/versions/3.0.3.md#securitySchemeObject
    """

    type: SecureType = Field(..., description="Secure scheme type")
    description: Optional[str] = Field(
        None,
        description="A short description for security scheme.",
    )
    name: Optional[str] = Field(
        None,
        description="The name of the header, query or cookie parameter to be used.",
    )
    field_in: Optional[InType] = Field(
        None, alias="in", description="The location of the API key."
    )
    scheme: Optional[str] = Field(
        None, description="The name of the HTTP Authorization scheme."
    )
    bearerFormat: Optional[str] = Field(
        None,
        description=(
            "A hint to the client to identify how the bearer token is formatted."
        ),
    )
    flows: Optional[dict] = Field(
        None,
        description=(
            "Containing configuration information for the flow types supported."
        ),
    )
    openIdConnectUrl: Optional[str] = Field(
        None, description="OpenId Connect URL to discover OAuth2 configuration values."
    )

    @model_validator(mode="before")
    @classmethod
    def check_type_required_fields(cls, values: dict):
        exist_fields = {key for key in values if values[key]}
        if not values.get("type"):
            raise ValueError("Type field is required")

        if not type_req_fields.get(values["type"], _EMPTY_SET).issubset(exist_fields):
            raise ValueError(
                f"For `{values['type']}` type "
                f"`{', '.join(type_req_fields[values['type']])}` field(s) is required. "
                f"But only found `{', '.join(exist_fields)}`."
            )
        return values

    model_config = ConfigDict(
        validate_assignment=True,
        validate_by_alias=True,
        validate_by_name=True,
    )


class SecurityScheme(BaseModel):
    """
    Named security scheme
    """

    name: str = Field(
        ...,
        description="Custom security scheme name. Can only contain - [A-Za-z0-9-._]",
    )
    data: SecuritySchemeData = Field(..., description="Security scheme data")

    @field_validator("name")
    def check_name(cls, value: str):
        if not OpenAPI_NAME_RE.fullmatch(value):
            raise ValueError("Name does not match OpenAPI rules")
        return value

    model_config = ConfigDict(validate_assignment=True)


class Server(BaseModel):
    """
    Servers section of OAS
    """

    url: str = Field(
        ...,
        description="""URL or path of API server

        (may be parametrized with using \"variables\" section - for more information,
        see: https://swagger.io/docs/specification/api-host-and-base-path/ )""",
    )
    description: Optional[str] = Field(
        None,
        description="Custom server description for server URL",
    )
    variables: Optional[dict] = Field(
        None,
        description="Variables for customizing server URL",
    )

    model_config = ConfigDict(validate_assignment=True)


class BaseFile:
    """
    An uploaded file, will be assigned as the corresponding web framework's
    file object.
    """

    @classmethod
    def __get_pydantic_json_schema__(cls, _core_schema: Dict[str, Any], _handler):
        return {"format": "binary", "type": "string"}

    @classmethod
    def __get_pydantic_core_schema__(cls, _source_type, _handler):
        return core_schema.with_info_plain_validator_function(cls.validate)

    @classmethod
    def validate(cls, value: Any, *_args, **_kwargs):
        return value



import json
from dataclasses import dataclass
from typing import Any, List

import pytest
from pydantic import BaseModel

from spectree._pydantic import (
    generate_root_model,
    is_base_model,
    is_base_model_instance,
    is_partial_base_model_instance,
    is_root_model,
    is_root_model_instance,
    serialize_model_instance,
)

DummyRootModel = generate_root_model(List[int], name="DummyRootModel")

NestedRootModel = generate_root_model(DummyRootModel, name="NestedRootModel")


class SimpleModel(BaseModel):
    user_id: int


Users = generate_root_model(List[SimpleModel], name="Users")


@dataclass
class RootModelLookalike:
    __root__: List[str]


@pytest.mark.parametrize(
    "value, expected",
    [
        (DummyRootModel, True),
        (DummyRootModel.model_validate([1, 2, 3]), False),
        (NestedRootModel, True),
        (
            NestedRootModel.model_validate(DummyRootModel.model_validate([1, 2, 3])),
            False,
        ),
        (SimpleModel, False),
        (SimpleModel(user_id=1), False),
        (RootModelLookalike, False),
        (RootModelLookalike(__root__=["False"]), False),
        (list, False),
        ([1, 2, 3], False),
        (str, False),
        ("str", False),
        (int, False),
        (1, False),
    ],
)
def test_is_root_model(value: Any, expected: bool):
    assert is_root_model(value) is expected


@pytest.mark.parametrize(
    "value, expected",
    [
        (DummyRootModel, False),
        (DummyRootModel.model_validate([1, 2, 3]), True),
        (NestedRootModel, False),
        (
            NestedRootModel.model_validate(DummyRootModel.model_validate([1, 2, 3])),
            True,
        ),
        (SimpleModel, False),
        (SimpleModel(user_id=1), False),
        (RootModelLookalike, False),
        (RootModelLookalike(__root__=["False"]), False),
        (list, False),
        ([1, 2, 3], False),
        (str, False),
        ("str", False),
        (int, False),
        (1, False),
    ],
)
def test_is_root_model_instance(value, expected):
    assert is_root_model_instance(value) is expected


@pytest.mark.parametrize(
    "value, expected",
    [
        (DummyRootModel, True),
        (DummyRootModel.model_validate([1, 2, 3]), False),
        (NestedRootModel, True),
        (
            NestedRootModel.model_validate(DummyRootModel.model_validate([1, 2, 3])),
            False,
        ),
        (SimpleModel, True),
        (SimpleModel(user_id=1), False),
        (RootModelLookalike, False),
        (RootModelLookalike(__root__=["False"]), False),
        (list, False),
        ([1, 2, 3], False),
        (str, False),
        ("str", False),
        (int, False),
        (1, False),
    ],
)
def test_is_base_model(value, expected):
    assert is_base_model(value) is expected


@pytest.mark.parametrize(
    "value, expected",
    [
        (DummyRootModel, False),
        (DummyRootModel.model_validate([1, 2, 3]), True),
        (NestedRootModel, False),
        (
            NestedRootModel.model_validate(DummyRootModel.model_validate([1, 2, 3])),
            True,
        ),
        (SimpleModel, False),
        (SimpleModel(user_id=1), True),
        (RootModelLookalike, False),
        (RootModelLookalike(__root__=["False"]), False),
        (list, False),
        ([1, 2, 3], False),
        (str, False),
        ("str", False),
        (int, False),
        (1, False),
    ],
)
def test_is_base_model_instance(value, expected):
    assert is_base_model_instance(value) is expected


@pytest.mark.parametrize(
    "value, expected",
    [
        (SimpleModel(user_id=1), True),
        ([0, SimpleModel(user_id=1)], True),
        ([1, 2, 3], False),
        ((0, SimpleModel(user_id=1)), True),
        ((0, 1), False),
        ({"test": SimpleModel(user_id=1)}, True),
        ({"test": [SimpleModel(user_id=1)]}, True),
        ([0, [1, SimpleModel(user_id=1)]], True),
    ],
)
def test_is_partial_base_model_instance(value, expected):
    assert is_partial_base_model_instance(value) is expected, value


@pytest.mark.parametrize(
    "value, expected",
    [
        (SimpleModel(user_id=1), {"user_id": 1}),
        (DummyRootModel.model_validate([1, 2, 3]), [1, 2, 3]),
        (
            NestedRootModel.model_validate(DummyRootModel.model_validate([1, 2, 3])),
            [1, 2, 3],
        ),
        (
            Users.model_validate(
                [
                    SimpleModel(user_id=1),
                    SimpleModel(user_id=2),
                ]
            ),
            [{"user_id": 1}, {"user_id": 2}],
        ),
    ],
)
def test_serialize_model_instance(value, expected):
    assert json.loads(serialize_model_instance(value).data) == expected



import inspect
from collections import namedtuple
from functools import partial
from json import JSONDecodeError
from typing import Any, Callable, Optional

from pydantic import ValidationError
from starlette.convertors import CONVERTOR_TYPES
from starlette.requests import Request
from starlette.responses import HTMLResponse, JSONResponse
from starlette.routing import compile_path

from spectree._pydantic import (
    SerializedPydanticResponse,
    generate_root_model,
    serialize_model_instance,
)
from spectree._types import ModelType
from spectree.plugins.base import (
    BasePlugin,
    Context,
    RawResponsePayload,
    validate_response,
)
from spectree.response import Response
from spectree.utils import cached_type_hints, get_multidict_items_starlette

METHODS = {"get", "post", "put", "patch", "delete"}
Route = namedtuple("Route", ["path", "methods", "func"])


_PydanticResponseModel = generate_root_model(Any, name="_PydanticResponseModel")


def PydanticResponse(content):
    class _PydanticResponse(JSONResponse):
        def render(self, content) -> bytes:
            self._model_class = content.__class__
            return serialize_model_instance(
                _PydanticResponseModel.model_validate(content)
            ).data

    return _PydanticResponse(content)


class StarlettePlugin(BasePlugin):
    ASYNC = True

    def __init__(self, spectree):
        super().__init__(spectree)

        self.conv2type = {conv: typ for typ, conv in CONVERTOR_TYPES.items()}

    def register_route(self, app):
        app.add_route(
            self.config.spec_url,
            lambda request: JSONResponse(self.spectree.spec),
        )

        for ui in self.config.page_templates:
            app.add_route(
                f"/{self.config.path}/{ui}",
                lambda request, ui=ui: HTMLResponse(
                    self.config.page_templates[ui].format(
                        spec_url=self.config.filename,
                        spec_path=self.config.path,
                        **self.config.swagger_oauth2_config(),
                    )
                ),
            )

    async def request_validation(self, request, query, json, form, headers, cookies):
        has_data = request.method not in ("GET", "DELETE")
        content_type = request.headers.get("content-type", "").lower()
        use_json = json and has_data and content_type == "application/json"
        use_form = (
            form and has_data and any([x in content_type for x in self.FORM_MIMETYPE])
        )
        request.context = Context(
            query.model_validate(get_multidict_items_starlette(request.query_params))
            if query
            else None,
            json.model_validate(await request.json() or {}) if use_json else None,
            form.model_validate(await request.form() or {}) if use_form else None,
            headers.model_validate(request.headers) if headers else None,
            cookies.model_validate(request.cookies) if cookies else None,
        )

    async def validate(
        self,
        func: Callable,
        query: Optional[ModelType],
        json: Optional[ModelType],
        form: Optional[ModelType],
        headers: Optional[ModelType],
        cookies: Optional[ModelType],
        resp: Optional[Response],
        before: Callable,
        after: Callable,
        validation_error_status: int,
        skip_validation: bool,
        force_resp_serialize: bool,
        *args: Any,
        **kwargs: Any,
    ):
        if isinstance(args[0], Request):
            instance, request = None, args[0]
        else:
            instance, request = args[:2]

        response = None
        req_validation_error = resp_validation_error = json_decode_error = None

        if not skip_validation:
            try:
                await self.request_validation(
                    request, query, json, form, headers, cookies
                )
            except ValidationError as err:
                req_validation_error = err
                response = JSONResponse(
                    err.errors(include_context=False), validation_error_status
                )
            except JSONDecodeError as err:
                json_decode_error = err
                self.logger.info(
                    "%s Validation Error",
                    validation_error_status,
                    extra={"spectree_json_decode_error": str(err)},
                )
                response = JSONResponse(
                    {"error_msg": str(err)}, validation_error_status
                )

        before(request, response, req_validation_error, instance)
        if req_validation_error or json_decode_error:
            return response

        if self.config.annotations:
            annotations = cached_type_hints(func)
            for name in ("query", "json", "form", "headers", "cookies"):
                if annotations.get(name):
                    kwargs[name] = getattr(
                        getattr(request, "context", None), name, None
                    )

        if inspect.iscoroutinefunction(func):
            response = await func(*args, **kwargs)
        else:
            response = func(*args, **kwargs)

        if (
            not skip_validation
            and resp
            and response
            and not (
                isinstance(response, JSONResponse)
                and hasattr(response, "_model_class")
                and response._model_class == resp.find_model(response.status_code)
            )
        ):
            try:
                response_validation_result = validate_response(
                    validation_model=resp.find_model(response.status_code),
                    response_payload=RawResponsePayload(payload=response.body),
                    force_serialize=force_resp_serialize,
                )
            except ValidationError as err:
                response = JSONResponse(
                    err.errors(include_context=False),
                    500,
                )
                resp_validation_error = err
            else:
                # replace the body of the response if it was serialized during validation
                if isinstance(
                    response_validation_result.payload, SerializedPydanticResponse
                ):
                    response.body = response_validation_result.payload.data

        after(request, response, resp_validation_error, instance)

        return response

    def find_routes(self):
        routes = []

        def parse_route(app, prefix=""):
            # :class:`starlette.staticfiles.StaticFiles` doesn't have routes
            if not app.routes:
                return
            for route in app.routes:
                if route.path.startswith(f"/{self.config.path}"):
                    continue

                func = route.app
                if isinstance(func, partial):
                    try:
                        func = func.__wrapped__
                    except AttributeError as err:
                        self.logger.warning(
                            "failed to get the wrapped func %s: %s", func, err
                        )

                if inspect.isclass(func):
                    for method in METHODS:
                        if getattr(func, method, None):
                            routes.append(
                                Route(
                                    f"{prefix}{route.path}",
                                    {method.upper()},
                                    getattr(func, method),
                                )
                            )
                elif inspect.isfunction(func):
                    routes.append(
                        Route(f"{prefix}{route.path}", route.methods, route.endpoint)
                    )
                else:
                    parse_route(route, prefix=f"{prefix}{route.path}")

        parse_route(self.spectree.app)
        return routes

    def bypass(self, func, method):
        return method in ["HEAD", "OPTIONS"]

    def parse_func(self, route):
        for method in route.methods or ["GET"]:
            yield method, route.func

    def parse_path(self, route, path_parameter_descriptions):
        _, path, variables = compile_path(route.path)
        parameters = []

        for name, conv in variables.items():
            schema = None
            typ = self.conv2type[conv]
            if typ == "int":
                schema = {"type": "integer", "format": "int32"}
            elif typ == "float":
                schema = {
                    "type": "number",
                    "format": "float",
                }
            elif typ == "path":
                schema = {
                    "type": "string",
                    "format": "path",
                }
            elif typ == "str":
                schema = {"type": "string"}

            description = (
                path_parameter_descriptions.get(name, "")
                if path_parameter_descriptions
                else ""
            )
            parameters.append(
                {
                    "name": name,
                    "in": "path",
                    "required": True,
                    "schema": schema,
                    "description": description,
                }
            )

        return path, parameters



import pytest
from falcon import App as FalconApp
from flask import Flask
from pydantic import BaseModel
from starlette.applications import Starlette

from spectree import Response
from spectree.config import Configuration
from spectree.models import Server, ValidationError
from spectree.plugins.flask_plugin import FlaskPlugin
from spectree.spec import SpecTree

from .common import get_paths


def backend_app():
    return [
        ("flask", Flask(__name__)),
        ("falcon", FalconApp()),
        ("starlette", Starlette()),
    ]


def _get_spec(name, app, **kwargs):
    api = SpecTree(name, app=app, title=f"{name}", **kwargs)
    if name == "flask":
        with app.app_context():
            spec = api.spec
    else:
        spec = api.spec

    return spec


def test_spectree_init():
    spec = SpecTree(path="docs")
    conf = Configuration()

    assert spec.config.title == conf.title
    assert spec.config.path == "docs"

    with pytest.raises(NotImplementedError):
        SpecTree(app=conf)


@pytest.mark.parametrize("name, app", backend_app())
def test_register(name, app):
    api = SpecTree(name)
    api.register(app)


@pytest.mark.parametrize("name, app", backend_app())
def test_spec_generate(name, app):
    spec = _get_spec(name, app)

    assert spec["info"]["title"] == name
    assert spec["paths"] == {}


@pytest.mark.parametrize("name, app", backend_app())
def test_spec_servers_empty(name, app):
    spec = _get_spec(name, app)

    assert "servers" not in spec


@pytest.mark.parametrize("name, app", backend_app())
def test_spec_servers_only(name, app):
    server1_url = "http://foo/bar"
    server2_url = "/foo/bar/"
    spec = _get_spec(
        name, app, servers=[Server(url=server1_url), Server(url=server2_url)]
    )

    assert spec["servers"] == [
        {"url": server1_url},
        {"url": server2_url},
    ]


@pytest.mark.parametrize("name, app", backend_app())
def test_spec_servers_full(name, app):
    server1 = {"url": "http://foo/bar", "description": "Foo Bar"}
    server2 = {"url": "http://bar/foo/{lang}", "variables": {"lang": "en"}}
    spec = _get_spec(
        name,
        app,
        servers=[
            Server(**server1),
            Server(**server2),
        ],
    )

    expected = []
    for server in [server1, server2]:
        expected_item = {
            "url": server.get("url"),
        }
        description = server.get("description", None)
        if description:
            expected_item["description"] = description
        variables = server.get("variables", None)
        if variables:
            expected_item["variables"] = variables
        expected.append(expected_item)

    assert spec["servers"] == expected


api = SpecTree("flask")
api_strict = SpecTree("flask", mode="strict")
api_greedy = SpecTree("flask", mode="greedy")
api_customize_backend = SpecTree(backend=FlaskPlugin)


def create_app():
    app = Flask(__name__)

    @app.route("/foo")
    @api.validate()
    def foo():
        pass

    @app.route("/bar")
    @api_strict.validate()
    def bar():
        pass

    @app.route("/lone", methods=["GET"])
    def lone_get():
        pass

    @app.route("/lone", methods=["POST"])
    def lone_post():
        pass

    return app


def test_spec_bypass_mode():
    app = create_app()
    api.register(app)
    with app.app_context():
        assert get_paths(api.spec) == ["/foo", "/lone"]

    app = create_app()
    api_customize_backend.register(app)
    with app.app_context():
        assert get_paths(api.spec) == ["/foo", "/lone"]

    app = create_app()
    api_greedy.register(app)
    with app.app_context():
        assert get_paths(api_greedy.spec) == ["/bar", "/foo", "/lone"]

    app = create_app()
    api_strict.register(app)
    with app.app_context():
        assert get_paths(api_strict.spec) == ["/bar"]


def test_two_endpoints_with_the_same_path():
    app = create_app()
    api.register(app)
    with app.app_context():
        spec = api.spec

    http_methods = list(spec["paths"]["/lone"].keys())
    http_methods.sort()
    assert http_methods == ["get", "post"]


def test_model_for_validation_errors_specified():
    api = SpecTree("flask")
    app = Flask(__name__)

    class CustomValidationError(BaseModel):
        pass

    @app.route("/foo")
    @api.validate(resp=Response(HTTP_200=None))
    def foo():
        pass

    @app.route("/bar")
    @api.validate(resp=Response(HTTP_200=None, HTTP_422=CustomValidationError))
    def bar():
        pass

    api.register(app)

    assert foo.resp.find_model(422) is ValidationError
    assert bar.resp.find_model(422) is CustomValidationError


def test_global_model_for_validation_errors_specified():
    class GlobalValidationError(BaseModel):
        pass

    class RouteValidationError(BaseModel):
        pass

    api = SpecTree("flask", validation_error_model=GlobalValidationError)
    app = Flask(__name__)

    @app.route("/foo")
    @api.validate(resp=Response(HTTP_200=None))
    def foo():
        pass

    @app.route("/bar")
    @api.validate(resp=Response(HTTP_200=None, HTTP_422=RouteValidationError))
    def bar():
        pass

    api.register(app)

    assert foo.resp.find_model(422) is GlobalValidationError
    assert bar.resp.find_model(422) is RouteValidationError


@pytest.mark.parametrize(
    ["override_operation_id", "expected_operation_id"],
    [(None, "get__foo"), ("getFoo", "getFoo")],
)
def test_operation_id_override(override_operation_id, expected_operation_id):
    api = SpecTree("flask")
    app = Flask(__name__)

    @app.route("/foo")
    @api.validate(operation_id=override_operation_id)
    def foo():
        pass

    api.register(app)

    with app.app_context():
        operation_id = api.spec["paths"]["/foo"]["get"]["operationId"]
        assert operation_id == expected_operation_id



from random import randint
from typing import List

import flask
import pytest
from flask import Blueprint, Flask, jsonify, make_response, request

from spectree import Response, SpecTree

from .common import (
    JSON,
    Cookies,
    CustomError,
    Form,
    FormFileUpload,
    Headers,
    ListJSON,
    OptionalAliasResp,
    Order,
    Query,
    QueryList,
    Resp,
    RespFromAttrs,
    RespObject,
    RootResp,
    StrDict,
    UserXmlData,
    api_tag,
    get_paths,
    get_root_resp_data,
)

# import tests to execute
from .flask_imports import *  # NOQA


def before_handler(req, resp, err, _):
    if err:
        resp.headers["X-Error"] = "Validation Error"


def after_handler(req, resp, err, _):
    resp.headers["X-Validation"] = "Pass"


def api_after_handler(req, resp, err, _):
    resp.headers["X-API"] = "OK"


api = SpecTree("flask", before=before_handler, after=after_handler, annotations=True)
app = Blueprint("test_blueprint", __name__)


@app.route("/ping")
@api.validate(headers=Headers, resp=Response(HTTP_202=StrDict), tags=["test", "health"])
def ping():
    """summary

    description"""
    return jsonify(msg="pong"), 202, request.context.headers.model_dump()


@app.route("/api/file_upload", methods=["POST"])
@api.validate(
    form=FormFileUpload,
)
def file_upload():
    upload = request.context.form.file
    assert upload
    return {
        "content": upload.stream.read().decode("utf-8"),
        "other": request.context.form.other,
    }


@app.route("/api/user/<name>", methods=["POST"])
@api.validate(
    query=Query,
    json=JSON,
    cookies=Cookies,
    form=Form,
    resp=Response(HTTP_200=Resp, HTTP_401=None),
    tags=[api_tag, "test"],
    after=api_after_handler,
)
def user_score(name):
    data_src = request.context.json or request.context.form
    score = [randint(0, int(data_src.limit)) for _ in range(5)]
    score.sort(reverse=request.context.query.order)
    assert request.context.cookies.pub == "abcdefg"
    assert request.cookies["pub"] == "abcdefg"
    return jsonify(name=data_src.name, score=score)


@app.route("/api/user_annotated/<name>", methods=["POST"])
@api.validate(
    resp=Response(HTTP_200=Resp, HTTP_401=None),
    tags=[api_tag, "test"],
    after=api_after_handler,
)
def user_score_annotated(name, query: Query, json: JSON, cookies: Cookies, form: Form):
    data_src = json or form
    score = [randint(0, int(data_src.limit)) for _ in range(5)]
    score.sort(reverse=(query.order == Order.desc))
    assert cookies.pub == "abcdefg"
    assert request.cookies["pub"] == "abcdefg"
    return jsonify(name=data_src.name, score=score)


@app.route("/api/user_skip/<name>", methods=["POST"])
@api.validate(
    query=Query,
    json=JSON,
    cookies=Cookies,
    resp=Response(HTTP_200=Resp, HTTP_401=None),
    tags=[api_tag, "test"],
    after=api_after_handler,
    skip_validation=True,
)
def user_score_skip_validation(name):
    response_format = request.args.get("response_format")
    assert response_format in ("json", "xml")
    json = request.get_json()
    score = [randint(0, json.get("limit")) for _ in range(5)]
    score.sort(reverse=int(request.args.get("order")) == Order.desc)
    assert request.cookies["pub"] == "abcdefg"
    if response_format == "json":
        return jsonify(name=name, x_score=score)
    else:
        return flask.Response(
            UserXmlData(name=name, score=score).dump_xml(),
            content_type="text/xml",
        )


@app.route("/api/user_model/<name>", methods=["POST"])
@api.validate(
    query=Query,
    json=JSON,
    cookies=Cookies,
    resp=Response(HTTP_200=Resp, HTTP_401=None),
    tags=[api_tag, "test"],
    after=api_after_handler,
)
def user_score_model(name):
    score = [randint(0, request.context.json.limit) for _ in range(5)]
    score.sort(reverse=request.context.query.order == Order.desc)
    assert request.context.cookies.pub == "abcdefg"
    assert request.cookies["pub"] == "abcdefg"
    return Resp(name=request.context.json.name, score=score)


@app.route("/api/user/<name>/address/<address_id>", methods=["GET"])
@api.validate(
    query=Query,
    path_parameter_descriptions={
        "name": "The name that uniquely identifies the user.",
        "non-existent-param": "description",
    },
)
def user_address(name, address_id):
    return None


@app.route("/api/no_response", methods=["GET", "POST"])
@api.validate(
    json=StrDict,
)
def no_response():
    return {}


@app.route("/api/set_cookies", methods=["GET"])
@api.validate(resp=Response(HTTP_200=StrDict))
def set_cookies():
    # related to GitHub issue #415
    resp = make_response(jsonify(msg="ping"))
    resp.set_cookie("foo", "hello")
    resp.set_cookie("bar", "world")
    return resp


@app.route("/api/list_json", methods=["POST"])
@api.validate(
    json=ListJSON,
)
def list_json():
    return {}


@app.route("/api/query_list")
@api.validate(query=QueryList)
def query_list():
    assert request.context.query.ids == [1, 2, 3]
    return {}


@app.route("/api/return_list", methods=["GET"])
@api.validate(resp=Response(HTTP_200=List[JSON]))
def return_list():
    pre_serialize = bool(int(request.args.get("pre_serialize", default=0)))
    data = [JSON(name="user1", limit=1), JSON(name="user2", limit=2)]
    return [entry.model_dump() if pre_serialize else entry for entry in data]


@app.route("/api/return_make_response", methods=["POST"])
@api.validate(json=JSON, headers=Headers, resp=Response(HTTP_201=Resp))
def return_make_response_post():
    model_data = request.context.json
    headers = request.context.headers
    response = make_response(
        Resp(name=model_data.name, score=[model_data.limit]).model_dump(), 201, headers
    )
    response.set_cookie(
        key="test_cookie",
        value=model_data.name,
        secure=True,
        httponly=True,
        samesite="Strict",
    )
    return response


@app.route("/api/return_make_response", methods=["GET"])
@api.validate(query=JSON, headers=Headers, resp=Response(HTTP_201=Resp))
def return_make_response_get():
    model_data = request.context.query
    headers = request.context.headers
    response = make_response(
        Resp(name=model_data.name, score=[model_data.limit]).model_dump(), 201, headers
    )
    response.set_cookie(
        key="test_cookie",
        value=model_data.name,
        secure=True,
        httponly=True,
        samesite="Strict",
    )
    return response


@app.route("/api/return_root", methods=["GET"])
@api.validate(resp=Response(HTTP_200=RootResp))
def return_root():
    return get_root_resp_data(
        pre_serialize=bool(int(request.args.get("pre_serialize", default=0))),
        return_what=request.args.get("return_what", default="RootResp"),
    )


@app.route("/api/return_model", methods=["GET"])
@api.validate()
def return_model():
    return get_root_resp_data(
        pre_serialize=False,
        return_what=request.args.get("return_what", default="RootResp"),
    )


@app.route("/api/return_string_status", methods=["GET"])
@api.validate()
def return_string_status():
    return "Response text string", 200


@app.route("/api/return_optional_alias", methods=["GET"])
@api.validate(resp=Response(HTTP_200=OptionalAliasResp))
def return_optional_alias():
    return {"schema": "test"}


@app.route("/api/custom_error", methods=["POST"])
@api.validate(resp=Response(HTTP_200=CustomError))
def custom_error(json: CustomError):
    return {"foo": "bar"}


@app.route("/api/force_serialize", methods=["GET"])
@api.validate(resp=Response(HTTP_200=RespFromAttrs), force_resp_serialize=True)
def force_serialize():
    return RespObject(name="flask", score=[1, 2, 3], comment="hello")


api.register(app)

flask_app = Flask(__name__)
flask_app.config["DEBUG"] = True
flask_app.config["TESTING"] = True
flask_app.register_blueprint(app)
with flask_app.app_context():
    _ = api.spec


@pytest.fixture
def client(request):
    parent_app = Flask(__name__)
    url_prefix = getattr(request, "param", None)
    parent_app.register_blueprint(app, url_prefix=url_prefix)
    with parent_app.test_client() as client:
        yield client


@pytest.mark.parametrize(
    ("client", "prefix"), [(None, ""), ("/prefix", "/prefix")], indirect=["client"]
)
def test_blueprint_prefix(client, prefix):
    resp = client.get(prefix + "/ping")
    assert resp.status_code == 422
    assert resp.headers.get("X-Error") == "Validation Error"

    resp = client.get(prefix + "/ping", headers={"lang": "en-US"})
    assert resp.status_code == 202, resp.text
    assert resp.json == {"msg": "pong"}
    assert resp.headers.get("X-Error") is None
    assert resp.headers.get("X-Validation") == "Pass"
    assert resp.headers.get("lang") == "en-US"


@pytest.fixture
def test_client_and_api(request):
    api_args = ["flask"]
    api_kwargs = {}
    endpoint_kwargs = {
        "headers": Headers,
        "resp": Response(HTTP_200=StrDict),
        "tags": ["test", "health"],
    }
    register_blueprint_kwargs = {}
    if hasattr(request, "param"):
        api_args.extend(request.param.get("api_args", ()))
        api_kwargs.update(request.param.get("api_kwargs", {}))
        endpoint_kwargs.update(request.param.get("endpoint_kwargs", {}))
        register_blueprint_kwargs.update(
            request.param.get("register_blueprint_kwargs", {})
        )

    api = SpecTree(*api_args, **api_kwargs)
    app = Blueprint("test_blueprint", __name__)

    @app.route("/ping")
    @api.validate(**endpoint_kwargs)
    def ping():
        """summary

        description"""
        return jsonify(msg="pong")

    api.register(app)

    flask_app = Flask(__name__)
    flask_app.register_blueprint(app, **register_blueprint_kwargs)

    with flask_app.app_context():
        _ = api.spec

    with flask_app.test_client() as test_client:
        yield test_client, api


@pytest.mark.parametrize(
    ("test_client_and_api", "prefix"),
    [
        ({"register_blueprint_kwargs": {}}, ""),
        ({"register_blueprint_kwargs": {"url_prefix": "/prefix"}}, "/prefix"),
    ],
    indirect=["test_client_and_api"],
)
def test_flask_doc_prefix(test_client_and_api, prefix):
    client, api = test_client_and_api

    resp = client.get(prefix + "/apidoc/openapi.json")
    assert resp.json == api.spec

    resp = client.get(prefix + "/apidoc/redoc/")
    assert resp.status_code == 200

    resp = client.get(prefix + "/apidoc/swagger/")
    assert resp.status_code == 200

    assert get_paths(api.spec) == [
        prefix + "/ping",
    ]



from dataclasses import dataclass
from typing import Any, Union, cast

from pydantic import BaseModel, RootModel

__all__ = [
    "BaseModel",
    "generate_root_model",
    "is_base_model",
    "is_base_model_instance",
    "is_pydantic_model",
    "is_root_model",
    "is_root_model_instance",
    "serialize_model_instance",
]


def generate_root_model(root_type, name="GeneratedRootModel") -> type[BaseModel]:
    return type(name, (RootModel[root_type],), {})


def is_pydantic_model(t: Any) -> bool:
    return issubclass(t, BaseModel)


def is_base_model(t: Any) -> bool:
    """Check whether a type is a Pydantic BaseModel"""
    try:
        return is_pydantic_model(t)
    except TypeError:
        return False


def is_base_model_instance(value: Any) -> bool:
    """Check whether a value is a Pydantic BaseModel instance."""
    return is_base_model(type(value))


def is_partial_base_model_instance(instance: Any) -> bool:
    """Check if it's a Pydantic BaseModel instance or [BaseModel]
    or {key: BaseModel} instance.
    """
    if not instance:
        return False
    if is_base_model_instance(instance):
        return True
    if isinstance(instance, dict):
        return any(
            is_partial_base_model_instance(key) or is_partial_base_model_instance(value)
            for key, value in instance.items()
        )
    if isinstance(instance, (list, tuple)):
        return any(is_partial_base_model_instance(value) for value in instance)
    return False


def is_root_model(t: Any) -> bool:
    """Check whether a type is a Pydantic RootModel."""
    pydantic_v2_root = is_base_model(t) and any(
        f"{m.__module__}.{m.__name__}" == "pydantic.root_model.RootModel"
        for m in t.mro()
    )
    return pydantic_v2_root


def is_root_model_instance(value: Any):
    """Check whether a value is a Pydantic RootModel instance."""
    return is_root_model(type(value))


@dataclass(frozen=True)
class SerializedPydanticResponse:
    data: bytes


_PydanticResponseModel = generate_root_model(Any, name="_PydanticResponseModel")


def serialize_model_instance(
    value: Union[BaseModel, list[BaseModel], dict[Any, BaseModel]],
) -> SerializedPydanticResponse:
    """Serialize a (partial) Pydantic BaseModel to json string."""
    if not is_base_model_instance(value):
        value = _PydanticResponseModel.model_validate(value)
    else:
        value = cast(BaseModel, value)
    serialized = value.model_dump_json()
    return SerializedPydanticResponse(serialized.encode("utf-8"))



from collections import namedtuple

from .base import BasePlugin

Plugin = namedtuple("Plugin", ("name", "package", "class_name"))

PLUGINS = {
    "base": Plugin(".base", __name__, "BasePlugin"),
    "flask": Plugin(".flask_plugin", __name__, "FlaskPlugin"),
    "quart": Plugin(".quart_plugin", __name__, "QuartPlugin"),
    "falcon": Plugin(".falcon_plugin", __name__, "FalconPlugin"),
    "falcon-asgi": Plugin(".falcon_plugin", __name__, "FalconAsgiPlugin"),
    "starlette": Plugin(".starlette_plugin", __name__, "StarlettePlugin"),
}

__all__ = ["PLUGINS", "BasePlugin", "Plugin"]
